@@ -22,8 +22,10 @@
 
 #include <Accounts/Account>
 #include <Accounts/AccountService>
+#include <Accounts/Application>
 #include <Accounts/Manager>
 #include <QPointer>
+#include <QQmlEngine>
 
 using namespace OnlineAccounts;
 
@@ -84,12 +86,14 @@ private:
     bool updateQueued;
     bool accountIdChanged;
     bool accountChanged;
+    bool applicationIdChanged;
     bool providerChanged;
     bool serviceTypeChanged;
     bool serviceChanged;
     bool includeDisabled;
     Accounts::AccountId accountId;
     QPointer<Accounts::Account> account;
+    Accounts::Application application;
     QString providerId;
     QString serviceTypeId;
     QString serviceId;
@@ -146,6 +150,8 @@ AccountServiceModelPrivate::listAccountServices(Accounts::Account *account) cons
     } else {
         foreach (Accounts::Service service, account->services()) {
             if (!serviceId.isEmpty() && service.name() != serviceId) continue;
+            if (application.isValid() &&
+                application.serviceUsage(service).isEmpty()) continue;
             ret.append(new Accounts::AccountService(account, service));
         }
     }
@@ -575,6 +581,7 @@ AccountServiceModel::AccountServiceModel(QObject *parent):
 
 AccountServiceModel::~AccountServiceModel()
 {
+    delete d_ptr;
 }
 
 void AccountServiceModel::classBegin()
@@ -633,6 +640,34 @@ QObject *AccountServiceModel::account() const
     return d->account;
 }
 
+/*!
+ * \qmlproperty string AccountServiceModel::applicationId
+ * If set, the model will only show those account services which are relevant
+ * for the given \a applicationId. This means that an account service will only
+ * be shown if it can be used by the application, as described in the
+ * application's manifest file.
+ */
+void AccountServiceModel::setApplicationId(const QString &applicationId)
+{
+    Q_D(AccountServiceModel);
+
+    if (applicationId == d->application.name()) return;
+    if (applicationId.isEmpty()) {
+        d->application = Accounts::Application();
+    } else {
+        d->application = SharedManager::instance()->application(applicationId);
+    }
+    d->applicationIdChanged = true;
+    d->queueUpdate();
+    Q_EMIT applicationIdChanged();
+}
+
+QString AccountServiceModel::applicationId() const
+{
+    Q_D(const AccountServiceModel);
+    return d->application.name();
+}
+
 /*!
  * \qmlproperty string AccountServiceModel::provider
  * If set, the model will list only those accounts services provided by this provider.
@@ -749,6 +784,7 @@ QVariant AccountServiceModel::data(const QModelIndex &index, int role) const
 
     Accounts::AccountService *accountService = d->modelItems.at(index.row());
     QVariant ret;
+    QObject *object = 0;
 
     switch (role) {
     case Qt::DisplayRole:
@@ -775,7 +811,7 @@ QVariant AccountServiceModel::data(const QModelIndex &index, int role) const
     case AccountServiceRole:
         qWarning("accountService role is deprecated, use accountServiceHandle");
     case AccountServiceHandleRole:
-        ret = QVariant::fromValue<QObject*>(accountService);
+        object = accountService;
         break;
     case AccountIdRole:
         ret = accountService->account()->id();
@@ -783,10 +819,14 @@ QVariant AccountServiceModel::data(const QModelIndex &index, int role) const
     case AccountRole:
         qWarning("account role is deprecated, use accountHandle");
     case AccountHandleRole:
-        ret = QVariant::fromValue<QObject*>(accountService->account());
+        object = accountService->account();
         break;
     }
 
+    if (object) {
+        QQmlEngine::setObjectOwnership(object, QQmlEngine::CppOwnership);
+        ret = QVariant::fromValue<QObject*>(object);
+    }
     return ret;
 }
 
