@@ -21,6 +21,7 @@
 #include "application.h"
 
 #include <Accounts/Manager>
+#include <QQmlEngine>
 
 using namespace OnlineAccounts;
 
@@ -46,6 +47,8 @@ using namespace OnlineAccounts;
  *     service; this is set to a valid value only if the \l
  *     ApplicationModel::service property is set to a valid service ID.
  * \li \c application is the Application object
+ * \li \c translations, the localization domain for translating the
+ * \c serviceUsage field
  * \endlist
  */
 
@@ -127,8 +130,12 @@ QVariant ApplicationModel::data(const QModelIndex &index, int role) const
         ret = application->serviceUsage(m_service);
         break;
     case ApplicationRole:
+        QQmlEngine::setObjectOwnership(application, QQmlEngine::CppOwnership);
         ret = QVariant::fromValue<QObject*>(application);
         break;
+    case TranslationsRole:
+        ret = application->trCatalog();
+        break;
     }
 
     return ret;
@@ -143,6 +150,7 @@ QHash<int, QByteArray> ApplicationModel::roleNames() const
         roles[IconNameRole] = "iconName";
         roles[ServiceUsageRole] = "serviceUsage";
         roles[ApplicationRole] = "application";
+        roles[TranslationsRole] = "translations";
     }
     return roles;
 }
