@@ -29,6 +29,26 @@
 
 using namespace Accounts;
 
+static bool mapIsSubset(const QVariantMap &set, const QVariantMap &test)
+{
+    QMapIterator<QString, QVariant> it(set);
+    while (it.hasNext()) {
+        it.next();
+        if (QMetaType::Type(it.value().type()) == QMetaType::QVariantMap) {
+            if (!mapIsSubset(it.value().toMap(),
+                             test.value(it.key()).toMap())) {
+                return false;
+            }
+        } else if (test.value(it.key()) != it.value()) {
+            qDebug() << "Maps differ: expected" << it.value() <<
+                "but found" << test.value(it.key());
+            return false;
+        }
+    }
+
+    return true;
+}
+
 class PluginTest: public QObject
 {
     Q_OBJECT
@@ -47,7 +67,11 @@ private Q_SLOTS:
     void testProviderModelWithApplication();
     void testAccountService();
     void testAccountServiceUpdate();
+    void testAuthentication_data();
     void testAuthentication();
+    void testAuthenticationErrors_data();
+    void testAuthenticationErrors();
+    void testAuthenticationDeleted();
     void testAuthenticationCancel();
     void testAuthenticationWithCredentials();
     void testManagerCreate();
@@ -863,7 +887,170 @@ void PluginTest::testAccountServiceUpdate()
     delete qmlObject;
 }
 
+void PluginTest::testAuthentication_data()
+{
+    QTest::addColumn<QString>("params");
+    QTest::addColumn<QVariantMap>("expectedReply");
+
+    QVariantMap reply;
+
+    reply.insert("test", QString("OK"));
+    reply.insert("host", QString("coolmail.ex"));
+    QTest::newRow("success with params") <<
+        "{ \"test\": \"OK\" }" <<
+        reply;
+    reply.clear();
+
+    reply.insert("host", QString("coolmail.ex"));
+    QTest::newRow("success without params") <<
+        "" <<
+        reply;
+    reply.clear();
+}
+
 void PluginTest::testAuthentication()
+{
+    QFETCH(QString, params);
+    QFETCH(QVariantMap, expectedReply);
+
+    clearDb();
+
+    /* Create one account */
+    Manager *manager = new Manager(this);
+    Service coolMail = manager->service("coolmail");
+    Service coolShare = manager->service("coolshare");
+    Account *account1 = manager->createAccount("cool");
+    QVERIFY(account1 != 0);
+
+    account1->setEnabled(true);
+    account1->setDisplayName("CoolAccount");
+    account1->selectService(coolMail);
+    account1->setEnabled(true);
+    account1->selectService(coolShare);
+    account1->setEnabled(false);
+    account1->syncAndBlock();
+
+    AccountService *accountService1 = new AccountService(account1, coolMail);
+    QVERIFY(accountService1 != 0);
+
+    QQmlEngine engine;
+    engine.rootContext()->setContextProperty("accountService1", accountService1);
+    QQmlComponent component(&engine);
+    component.setData("import Ubuntu.OnlineAccounts 0.1\n"
+                      "AccountService {\n"
+                      "  objectHandle: accountService1\n"
+                      "  function go() {\n"
+                      "    authenticate(" + params.toUtf8() + ")\n"
+                      "  }\n"
+                      "}",
+                      QUrl());
+    QObject *qmlObject = component.create();
+    QVERIFY(qmlObject != 0);
+
+    QSignalSpy authenticated(qmlObject,
+                             SIGNAL(authenticated(const QVariantMap &)));
+    QSignalSpy authenticationError(qmlObject,
+        SIGNAL(authenticationError(const QVariantMap &)));
+
+    QMetaObject::invokeMethod(qmlObject, "go");
+
+    QTRY_COMPARE(authenticated.count(), 1);
+    QCOMPARE(authenticationError.count(), 0);
+    QVariantMap reply = authenticated.at(0).at(0).toMap();
+    QVERIFY(mapIsSubset(expectedReply, reply));
+
+    delete accountService1;
+    delete manager;
+    delete qmlObject;
+}
+
+void PluginTest::testAuthenticationErrors_data()
+{
+    QTest::addColumn<QString>("params");
+    QTest::addColumn<QString>("codeName");
+    QTest::addColumn<QString>("expectedMessage");
+
+    QTest::newRow("Signon::UserCanceled") <<
+        "{ \"errorCode\": 311, \"errorMessage\": \"Failed!\" }" <<
+        "UserCanceledError" << "Failed!";
+
+    QTest::newRow("Signon::InvalidQuery") <<
+        "{ \"errorCode\": 103, \"errorMessage\": \"Weird\" }" <<
+        "NoAccountError" << "Weird";
+
+    QTest::newRow("Signon::PermissionDenied") <<
+        "{ \"errorCode\": 4, \"errorMessage\": \"Failed!\" }" <<
+        "PermissionDeniedError" << "Failed!";
+
+    QTest::newRow("Signon::NoConnection") <<
+        "{ \"errorCode\": 307, \"errorMessage\": \"Failed!\" }" <<
+        "NetworkError" << "Failed!";
+}
+
+void PluginTest::testAuthenticationErrors()
+{
+    QFETCH(QString, params);
+    QFETCH(QString, codeName);
+    QFETCH(QString, expectedMessage);
+
+    clearDb();
+
+    /* Create one account */
+    Manager *manager = new Manager(this);
+    Service coolMail = manager->service("coolmail");
+    Service coolShare = manager->service("coolshare");
+    Account *account1 = manager->createAccount("cool");
+    QVERIFY(account1 != 0);
+
+    account1->setEnabled(true);
+    account1->setDisplayName("CoolAccount");
+    account1->selectService(coolMail);
+    account1->setEnabled(true);
+    account1->selectService(coolShare);
+    account1->setEnabled(false);
+    account1->syncAndBlock();
+
+    AccountService *accountService1 = new AccountService(account1, coolMail);
+    QVERIFY(accountService1 != 0);
+
+    QQmlEngine engine;
+    engine.rootContext()->setContextProperty("accountService1", accountService1);
+    QQmlComponent component(&engine);
+    component.setData("import Ubuntu.OnlineAccounts 0.1\n"
+                      "AccountService {\n"
+                      "  objectHandle: accountService1\n"
+                      "  function go() {\n"
+                      "    authenticate(" + params.toUtf8() + ")\n"
+                      "  }\n"
+                      "}",
+                      QUrl());
+    QObject *qmlObject = component.create();
+    qDebug() << component.errors();
+    QVERIFY(qmlObject != 0);
+
+    QSignalSpy authenticationError(qmlObject,
+        SIGNAL(authenticationError(const QVariantMap &)));
+
+    QMetaObject::invokeMethod(qmlObject, "go");
+
+    QTRY_COMPARE(authenticationError.count(), 1);
+
+    QVariantMap reply = authenticationError.at(0).at(0).toMap();
+    int code = reply.value("code").toInt();
+    QString message = reply.value("message").toString();
+
+    QCOMPARE(message, expectedMessage);
+    const QMetaObject *mo = qmlObject->metaObject();
+    const QMetaEnum errorEnum =
+        mo->enumerator(mo->indexOfEnumerator("ErrorCode"));
+    QCOMPARE(code, errorEnum.keyToValue(codeName.toUtf8().constData()));
+
+    delete accountService1;
+    delete manager;
+    delete qmlObject;
+}
+
+void PluginTest::testAuthenticationDeleted()
 {
     clearDb();
 
@@ -900,33 +1087,7 @@ void PluginTest::testAuthentication()
         SIGNAL(authenticationError(const QVariantMap &)));
 
     QVariantMap sessionData;
-    sessionData.insert("test", QString("OK"));
-    QMetaObject::invokeMethod(qmlObject, "authenticate",
-                              Q_ARG(QVariantMap, sessionData));
-    QTest::qWait(50);
-
-    QCOMPARE(authenticationError.count(), 0);
-    QCOMPARE(authenticated.count(), 1);
-    QVariantMap reply = authenticated.at(0).at(0).toMap();
-    QVERIFY(!reply.isEmpty());
-    QCOMPARE(reply["test"].toString(), QString("OK"));
-    QCOMPARE(reply["host"].toString(), QString("coolmail.ex"));
-    authenticated.clear();
-
-    /* Test an authentication failure */
-    sessionData.insert("errorCode", 123);
-    sessionData.insert("errorMessage", QString("Failed!"));
-    QMetaObject::invokeMethod(qmlObject, "authenticate",
-                              Q_ARG(QVariantMap, sessionData));
-    QTest::qWait(50);
-
-    QCOMPARE(authenticationError.count(), 1);
-    QCOMPARE(authenticated.count(), 0);
-    QVariantMap error = authenticationError.at(0).at(0).toMap();
-    QVERIFY(!error.isEmpty());
-    QCOMPARE(error["code"].toInt(), 123);
-    QCOMPARE(error["message"].toString(), QString("Failed!"));
-    authenticationError.clear();
+    QVariantMap error;
 
     /* Delete the account service, and check that the QML object survives */
     delete accountService1;
@@ -1011,9 +1172,10 @@ void PluginTest::testAuthenticationCancel()
     QCOMPARE(authenticationError.count(), 1);
     QVariantMap error = authenticationError.at(0).at(0).toMap();
     QVERIFY(!error.isEmpty());
-    /* 311 is the value of SignOn::Error::SessionCanceled; we prefer avoiding
-     * #including any headers from libsignon from within this file. */
-    QCOMPARE(error["code"].toInt(), 311);
+    const QMetaObject *mo = qmlObject->metaObject();
+    const QMetaEnum errorEnum =
+        mo->enumerator(mo->indexOfEnumerator("ErrorCode"));
+    QCOMPARE(error["code"].toInt(), errorEnum.keyToValue("UserCanceledError"));
     authenticationError.clear();
 
     delete accountService1;
